#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

#define NINTH (1.0/9.0)

static double ** A;
static double ** C;
int N;
int NUM_PROCS;

void matrix_blur(int);

int main(int argc, char ** argv)
{
  N = atoi(argv[1]);
  int rank;
  int npes;
  MPI_Request * req;
  MPI_Status * statuses;

  statuses = (MPI_Status *) malloc(N*sizeof(MPI_Status));
  req = (MPI_Request *) malloc(N*sizeof(MPI_Request));
  A = (double **) malloc(N*sizeof(double *));
  C = (double **) malloc(N*sizeof(double *));

  int i, j;

  for(i = 0; i < N; i++)
  {
    A[i] = (double *) malloc(N*sizeof(double));
    C[i] = (double *) malloc(N*sizeof(double));
  }

  for(i = 0; i < N; i++)
    for(j = 0; j < N; j++)
      A[i][j] = i;

  MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &npes);
    NUM_PROCS=npes-1;

    if(rank == 0)
    {
      for(i = 0; i < NUM_PROCS; i++)
        for(j = 0; j < N/NUM_PROCS; j++)
          MPI_Irecv(C[i*(N/NUM_PROCS)+j], N, MPI_DOUBLE, i+1, j,
            MPI_COMM_WORLD, &req[i*(N/NUM_PROCS)+j]);

      MPI_Waitall(N, req, statuses);

      for(i = 0; i < 2; i++)
      {
        for(j = 0; j < N; j++)
          printf("%lf ", C[i][j]);
        printf("\n");
      }
     printf("\n");

      for(i = N-2; i < N; i++)
      {
        for(j = 0; j < N; j++)
          printf("%lf ", C[i][j]);
        printf("\n");
      }
      printf("\n");
    }
    else
    {
      matrix_blur(rank-1);

      for(i = 0; i < N/NUM_PROCS; i++)
      {
        int loc = (rank-1)*(N/NUM_PROCS)+i;
        MPI_Isend(C[loc], N, MPI_DOUBLE, 0, i, MPI_COMM_WORLD, &req[loc]);
      }
    }

    for(i = 0; i < N; i++)
    {
      free((void *) A[i]);
      free((void *) C[i]);
    }

    free((void *) A);
    free((void *) C);

  MPI_Finalize();


  return 0;
}

void matrix_blur(int pid)
{
  int i, j;
  int workToDo = (N/NUM_PROCS);
  for(i = pid*workToDo; i < pid*workToDo + workToDo; i++)
  {
    for(j = 0; j < N; j++)
    {
      C[i][j] = NINTH*A[i][j];
      if(i-1 >= 0)
        C[i][j] += NINTH*A[i-1][j];
      if(i+1 < N)
        C[i][j] += NINTH*A[i+1][j];
      if(j-1 >= 0)
        C[i][j] += NINTH*A[i][j-1];
      if(j+1 < N)
        C[i][j] += NINTH*A[i][j+1];
      if(i-1 >= 0 && j-1 >= 0)
        C[i][j] += NINTH*A[i-1][j-1];
      if(i-1 >= 0 && j+1 < N)
        C[i][j] += NINTH*A[i-1][j+1];
      if(i+1 < N && j+1 < N)
        C[i][j] += NINTH*A[i+1][j+1];
      if(i+1 < N && j-1 >= 0)
        C[i][j] += NINTH*A[i+1][j-1];
    }
  }
}
