#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#define NINTH (1.0/9.0)

static double ** A;
static double ** C;
int N;
int NUM_THREADS;

void * matrix_blur(void *);

int main(int argc, char ** argv)
{
  printf("Please enter the matrix size (N) and the number of threads: ");
  scanf("%d %d", &N, &NUM_THREADS);

  pthread_t * thread_list = (pthread_t *) 
    malloc(NUM_THREADS*sizeof(pthread_t));
  
  A = (double **) malloc(N*sizeof(double *));
  C = (double **) malloc(N*sizeof(double *));

  int i, j;

  for(i = 0; i < N; i++)
  {
    A[i] = (double *) malloc(N*sizeof(double));
    C[i] = (double *) malloc(N*sizeof(double));
  }

  for(i = 0; i < N; i++)
    for(j = 0; j < N; j++)
      A[i][j] = i;

  for(i = 0; i < NUM_THREADS; i++)
    pthread_create(&thread_list[i], NULL, matrix_blur, (void *) i);

  for(i = 0; i < NUM_THREADS; i++)
    pthread_join(thread_list[i], NULL);

  for(i = 0; i < 2; i++)
  {
    for(j = 0; j < N; j++)
      printf("%lf ", C[i][j]);
    printf("\n");
  }
  printf("\n");

  for(i = N-2; i < N; i++)
  {
    for(j = 0; j < N; j++)
      printf("%lf ", C[i][j]);
    printf("\n");
  }
  printf("\n");

  for(i = 0; i < N; i++)
  {
    free((void *) A[i]);
    free((void *) C[i]);
  }

  free((void *) A);
  free((void *) C);
  return 0;
}

void * matrix_blur(void * threadId)
{
  int tid = (int) threadId;
  
  int i, j;
  int workToDo = (N/NUM_THREADS);
  for(i = tid*workToDo; i < tid*workToDo + workToDo; i++)
  {
    for(j = 0; j < N; j++)
    {
      C[i][j] = NINTH*A[i][j];
      if(i-1 >= 0)
        C[i][j] += NINTH*A[i-1][j];
      if(i+1 < N)
        C[i][j] += NINTH*A[i+1][j];
      if(j-1 >= 0)
        C[i][j] += NINTH*A[i][j-1];
      if(j+1 < N)
        C[i][j] += NINTH*A[i][j+1];
      if(i-1 >= 0 && j-1 >= 0)
        C[i][j] += NINTH*A[i-1][j-1];
      if(i-1 >= 0 && j+1 < N)
        C[i][j] += NINTH*A[i-1][j+1];
      if(i+1 < N && j+1 < N)
        C[i][j] += NINTH*A[i+1][j+1];
      if(i+1 < N && j-1 >= 0)
        C[i][j] += NINTH*A[i+1][j-1];
    }
  }
  pthread_exit(NULL);
}
